##
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
##
import os
import json

import logging
import warnings

logger = logging.getLogger(__name__)

from typing import Any, Dict, Tuple, Union, List, Optional
from azure.quantum.version import __version__
from azure.quantum.qiskit.job import (
    MICROSOFT_OUTPUT_DATA_FORMAT,
    MICROSOFT_OUTPUT_DATA_FORMAT_V2,
    AzureQuantumJob,
)
from abc import abstractmethod
from azure.quantum.job.session import SessionHost

try:
    from qiskit import QuantumCircuit
    from qiskit.providers import BackendV1 as Backend
    from qiskit.providers import Options
    from qiskit.providers import Provider
    from qiskit.providers.models import BackendConfiguration
    from qiskit.qobj import QasmQobj, PulseQobj
    import pyqir as pyqir
    from qsharp.interop.qiskit import QSharpBackend
    from qsharp import TargetProfile

except ImportError:
    raise ImportError(
        "Missing optional 'qiskit' dependencies. \
To install run: pip install azure-quantum[qiskit]"
    )

# barrier is handled by an extra flag which will transpile
# them away if the backend doesn't support them. This has
# to be done as a special pass as the transpiler will not
# remove barriers by default.
QIR_BASIS_GATES = [
    "measure",
    "reset",
    "ccx",
    "cx",
    "cy",
    "cz",
    "rx",
    "rxx",
    "crx",
    "ry",
    "ryy",
    "cry",
    "rz",
    "rzz",
    "crz",
    "h",
    "s",
    "sdg",
    "swap",
    "t",
    "tdg",
    "x",
    "y",
    "z",
    "id",
    "ch",
]


class AzureBackendBase(Backend, SessionHost):

    # Name of the provider's input parameter which specifies number of shots for a submitted job.
    # If None, backend will not pass this input parameter. 
    _SHOTS_PARAM_NAME = None

    @abstractmethod
    def __init__(
        self,
        configuration: BackendConfiguration,
        provider: Provider = None,
        **fields
    ):
        super().__init__(configuration, provider, **fields)
    
    @abstractmethod
    def run(
        self,
        run_input: Union[QuantumCircuit, List[QuantumCircuit]] = [],
        shots: int = None, 
        **options,
    ) -> AzureQuantumJob:
        """Run on the backend.

        This method returns a
        :class:`~azure.quantum.qiskit.job.AzureQuantumJob` object
        that runs circuits. 

        Args:
            run_input (QuantumCircuit or List[QuantumCircuit]): An individual or a
            list of :class:`~qiskit.circuits.QuantumCircuit` to run on the backend.
            shots (int, optional): Number of shots, defaults to None.
            options: Any kwarg options to pass to the backend for running the
            config. If a key is also present in the options
            attribute/object then the expectation is that the value
            specified will be used instead of what's set in the options
            object.

        Returns:
            Job: The job object for the run
        """
        pass

    @classmethod
    def _can_send_shots_input_param(cls) -> bool:
        """
        Tells if provider's backend class is able to specify shots number for its jobs.
        """
        return cls._SHOTS_PARAM_NAME is not None

    @classmethod
    @abstractmethod
    def _default_options(cls) -> Options:
        pass

    @abstractmethod
    def _azure_config(self) -> Dict[str, str]:
        pass

    def retrieve_job(self, job_id) -> AzureQuantumJob:
        """Returns the Job instance associated with the given id."""
        return self._provider.get_job(job_id)

    def _get_output_data_format(self, options: Dict[str, Any] = {}) -> str:
        config: BackendConfiguration = self.configuration()
        # output data format default depends on the number of experiments. QIR backends
        # that don't define a default in their azure config will use this value
        # Once more than one experiment is supported, we should always use the v2 format
        default_output_data_format = (
            MICROSOFT_OUTPUT_DATA_FORMAT
            if config.max_experiments == 1
            else MICROSOFT_OUTPUT_DATA_FORMAT_V2
        )

        azure_config: Dict[str, Any] = config.azure
        # if the backend defines an output format, use that over the default
        azure_defined_override = azure_config.get(
            "output_data_format", default_output_data_format
        )
        # if the user specifies an output format, use that over the default azure config
        output_data_format = options.pop("output_data_format", azure_defined_override)

        return output_data_format
    
    def _get_input_params(self, options: Dict[str, Any], shots: int = None) -> Dict[str, Any]:
        # Backend options are mapped to input_params.
        input_params: Dict[str, Any] = vars(self.options).copy()

        # Determine shots number, if needed.
        if self._can_send_shots_input_param():
            options_shots = options.pop(self.__class__._SHOTS_PARAM_NAME, None)

            final_shots = None
            # First we check for the explicitly specified 'shots' parameter, then for a provider-specific
            # field in options, then for a backend's default value. 

            # Warn about options conflict, default to 'shots'.
            if shots is not None and options_shots is not None:
                warnings.warn(
                    f"Parameter 'shots' conflicts with the '{self.__class__._SHOTS_PARAM_NAME}' parameter. "
                    "Please, provide only one option for setting shots. Defaulting to 'shots' parameter."
                )
                final_shots = shots
            
            elif shots is not None:
                final_shots = shots
            elif options_shots is not None:
                warnings.warn(
                    f"Parameter '{self.__class__._SHOTS_PARAM_NAME}' is subject to change in future versions. "
                    "Please, use 'shots' parameter instead."
                )
                final_shots = options_shots
            
            # If nothing is found, try to get from default values.
            if final_shots is None:
                final_shots = input_params.get(self.__class__._SHOTS_PARAM_NAME)

            # Also add all possible shots options into input_params to make sure 
            # that all backends covered. 
            # TODO: Double check all backends for shots options in order to remove this extra check.
            input_params["shots"] = final_shots
            input_params["count"] = final_shots

            # Safely removing "shots" and "count" from options as they will be passed in input_params now.
            _ = options.pop("shots", None)
            _ = options.pop("count", None)

            input_params[self.__class__._SHOTS_PARAM_NAME] = final_shots
            

        if "items" in options:
            input_params["items"] = options.pop("items")

        # Take also into consideration options passed in the options, as the take precedence
        # over default values:
        for opt in options.copy():
            if opt in input_params:
                input_params[opt] = options.pop(opt)

        return input_params

    def _run(self, job_name, input_data, input_params, metadata, **options):
        logger.info(f"Submitting new job for backend {self.name()}")

        # The default of these job parameters come from the AzureBackend configuration:
        config = self.configuration()
        blob_name = options.pop("blob_name", config.azure["blob_name"])
        content_type = options.pop("content_type", config.azure["content_type"])
        provider_id = options.pop("provider_id", config.azure["provider_id"])
        input_data_format = options.pop(
            "input_data_format", config.azure["input_data_format"]
        )
        output_data_format = self._get_output_data_format(options)

        # QIR backends will have popped "targetCapability" to configure QIR generation.
        # Anything left here is an invalid parameter with the user attempting to use
        # deprecated parameters.
        targetCapability = input_params.get("targetCapability", None)
        if (
            targetCapability not in [None, "qasm"]
            and input_data_format != "qir.v1"
        ):
            message = "The targetCapability parameter has been deprecated and is only supported for QIR backends."
            message += os.linesep
            message += "To find a QIR capable backend, use the following code:"
            message += os.linesep
            message += (
                f'\tprovider.get_backend("{self.name()}", input_data_format: "qir.v1").'
            )
            raise ValueError(message)

        job = AzureQuantumJob(
            backend=self,
            target=self.name(),
            name=job_name,
            input_data=input_data,
            blob_name=blob_name,
            content_type=content_type,
            provider_id=provider_id,
            input_data_format=input_data_format,
            output_data_format=output_data_format,
            input_params=input_params,
            metadata=metadata,
            **options,
        )

        return job

    def _normalize_run_input_params(self, run_input, **options):
        if "circuit" not in options:
            # circuit is not provided, check if there is run_input
            if run_input:
                return run_input
            else:
                raise ValueError("No input provided.")

        if run_input:
            # even though circuit is provided, we still have run_input
            warnings.warn(
                DeprecationWarning(
                    "The circuit parameter has been deprecated and will be ignored."
                )
            )
            return run_input
        else:
            warnings.warn(
                DeprecationWarning(
                    "The circuit parameter has been deprecated. Please use the run_input parameter."
                )
            )

        # we don't have run_input
        # we know we have circuit parameter, but it may be empty
        circuit = options.get("circuit")

        if circuit:
            return circuit
        else:
            raise ValueError("No input provided.")

    def _get_azure_workspace(self) -> "Workspace":
        return self.provider().get_workspace()

    def _get_azure_target_id(self) -> str:
        return self.name()

    def _get_azure_provider_id(self) -> str:
        return self._azure_config()["provider_id"]


class AzureQirBackend(AzureBackendBase):
    @abstractmethod
    def __init__(
        self, configuration: BackendConfiguration, provider: Provider = None, **fields
    ):
        super().__init__(configuration, provider, **fields)

    def _azure_config(self) -> Dict[str, str]:
        return {
            "blob_name": "inputData",
            "content_type": "qir.v1",
            "input_data_format": "qir.v1",
            "output_data_format": "microsoft.quantum-results.v2",
            "is_default": True,
        }
    
    def _basis_gates(self) -> List[str]:
        return QIR_BASIS_GATES

    def run(
        self,
        run_input: Union[QuantumCircuit, List[QuantumCircuit]] = [],
        shots: int = None, 
        **options,
    ) -> AzureQuantumJob:
        """Run on the backend.

        This method returns a
        :class:`~azure.quantum.qiskit.job.AzureQuantumJob` object
        that runs circuits.

        Args:
            run_input (QuantumCircuit or List[QuantumCircuit]): An individual or a
            list of :class:`~qiskit.circuits.QuantumCircuit` to run on the backend.
            shots (int, optional): Number of shots, defaults to None.
            options: Any kwarg options to pass to the backend for running the
            config. If a key is also present in the options
            attribute/object then the expectation is that the value
            specified will be used instead of what's set in the options
            object.

        Returns:
            Job: The job object for the run
        """
        run_input = self._normalize_run_input_params(run_input, **options)
        options.pop("run_input", None)
        options.pop("circuit", None)

        circuits = list([])
        if isinstance(run_input, QuantumCircuit):
            circuits = [run_input]
        else:
            circuits = run_input

        max_circuits_per_job = self.configuration().max_experiments
        if len(circuits) > max_circuits_per_job:
            raise NotImplementedError(
                f"This backend only supports running a maximum of {max_circuits_per_job} circuits per job."
            )

        # config normalization
        input_params = self._get_input_params(options, shots=shots)
        
        shots_count = None

        if self._can_send_shots_input_param():
            shots_count = input_params.get(self.__class__._SHOTS_PARAM_NAME)

        job_name = ""
        if len(circuits) > 1:
            job_name = f"batch-{len(circuits)}"
            if shots_count is not None:
                job_name = f"{job_name}-{shots_count}"
        else:
            job_name = circuits[0].name
        job_name = options.pop("job_name", job_name)

        metadata = options.pop("metadata", self._prepare_job_metadata(circuits))

        input_data = self._translate_input(circuits, input_params)

        job = super()._run(job_name, input_data, input_params, metadata, **options)
        logger.info(
            f"Submitted job with id '{job.id()}' with shot count of {shots_count}:"
        )

        return job

    def _prepare_job_metadata(self, circuits: List[QuantumCircuit]) -> Dict[str, str]:
        """Returns the metadata relative to the given circuits that will be attached to the Job"""
        if len(circuits) == 1:
            circuit: QuantumCircuit = circuits[0]
            return {
                "qiskit": str(True),
                "name": circuit.name,
                "num_qubits": circuit.num_qubits,
                "metadata": json.dumps(circuit.metadata),
            }
        # for batch jobs, we don't want to store the metadata of each circuit
        # we fill out the result header in output processing.
        # These headers don't matter for execution are are only used for
        # result processing.
        return {}

    def _generate_qir(
        self, circuits: List[QuantumCircuit], target_profile: TargetProfile, **kwargs
    ) -> pyqir.Module:

        if len(circuits) == 0:
            raise ValueError("No QuantumCircuits provided")

        config = self.configuration()
        # Barriers aren't removed by transpilation and must be explicitly removed in the Qiskit to QIR translation.
        supports_barrier = "barrier" in config.basis_gates
        skip_transpilation = kwargs.pop("skip_transpilation", False)

        backend = QSharpBackend(
            qiskit_pass_options={"supports_barrier": supports_barrier},
            target_profile=target_profile,
            skip_transpilation=skip_transpilation,
            **kwargs,
        )

        name = "batch"
        if len(circuits) == 1:
            name = circuits[0].name

        if isinstance(circuits, list):
            for value in circuits:
                if not isinstance(value, QuantumCircuit):
                    raise ValueError("Input must be List[QuantumCircuit]")
        else:
            raise ValueError("Input must be List[QuantumCircuit]")

        context = pyqir.Context()
        llvm_module = pyqir.qir_module(context, name)
        for circuit in circuits:
            qir_str = backend.qir(circuit)
            module = pyqir.Module.from_ir(context, qir_str)
            entry_point = next(filter(pyqir.is_entry_point, module.functions))
            entry_point.name = circuit.name
            llvm_module.link(module)
        err = llvm_module.verify()
        if err is not None:
            raise Exception(err)

        return llvm_module

    def _get_qir_str(
        self,
        circuits: List[QuantumCircuit],
        target_profile: TargetProfile,
        **to_qir_kwargs,
    ) -> str:
        module = self._generate_qir(circuits, target_profile, **to_qir_kwargs)
        return str(module)

    def _translate_input(
        self, circuits: Union[QuantumCircuit, List[QuantumCircuit]], input_params: Dict[str, Any]
    ) -> bytes:
        """Translates the input values to the QIR expected by the Backend."""
        logger.info(f"Using QIR as the job's payload format.")
        if not (isinstance(circuits, list)):
            circuits = [circuits]

        target_profile = self._get_target_profile(input_params)

        if logger.isEnabledFor(logging.DEBUG):
            qir = self._get_qir_str(circuits, target_profile, skip_transpilation=True)
            logger.debug(f"QIR:\n{qir}")

        # We'll transpile automatically to the supported gates in QIR unless explicitly skipped.
        skip_transpilation = input_params.pop("skipTranspile", False)

        module = self._generate_qir(
            circuits, target_profile, skip_transpilation=skip_transpilation
        )

        def get_func_name(func: pyqir.Function) -> str:
            return func.name

        entry_points = list(
            map(get_func_name, filter(pyqir.is_entry_point, module.functions))
        )

        if not skip_transpilation:
            # We'll only log the QIR again if we performed a transpilation.
            if logger.isEnabledFor(logging.DEBUG):
                qir = str(module)
                logger.debug(f"QIR (Post-transpilation):\n{qir}")

        if "items" not in input_params:
            arguments = input_params.pop("arguments", [])
            input_params["items"] = [
                {"entryPoint": name, "arguments": arguments} for name in entry_points
            ]

        return str(module).encode("utf-8")

    def _estimate_cost_qir(
        self, circuits: Union[QuantumCircuit, List[QuantumCircuit]], shots, options={}
    ):
        """Estimate the cost for the given circuit."""
        input_params = self._get_input_params(options, shots=shots)

        if not (isinstance(circuits, list)):
            circuits = [circuits]

        skip_transpilation = input_params.pop("skipTranspile", False)
        target_profile = self._get_target_profile(input_params)
        module = self._generate_qir(
            circuits, target_profile, skip_transpilation=skip_transpilation
        )

        workspace = self.provider().get_workspace()
        target = workspace.get_targets(self.name())
        return target.estimate_cost(module, shots=shots)

    def _get_target_profile(self, input_params) -> TargetProfile:
        # Default to Adaptive_RI if not specified on the backend
        # this is really just a safeguard in case the backend doesn't have a default
        default_profile = self.options.get("target_profile", TargetProfile.Adaptive_RI)

        # If the user is using the old targetCapability parameter, we'll warn them
        # and use that value for now. This will be removed in the future.
        if "targetCapability" in input_params:
            warnings.warn(
                "The 'targetCapability' parameter is deprecated and will be ignored in the future. "
                "Please, use 'target_profile' parameter instead.",
                category=DeprecationWarning,
            )
            cap = input_params.pop("targetCapability")
            if cap == "AdaptiveExecution":
                default_profile = TargetProfile.Adaptive_RI
            else:
                default_profile = TargetProfile.Base
        # If the user specifies a target profile, use that.
        # Otherwise, use the profile we got from the backend/targetCapability.
        return input_params.pop("target_profile", default_profile)


class AzureBackend(AzureBackendBase):
    """Base class for interfacing with a backend in Azure Quantum"""

    @abstractmethod
    def __init__(
        self, configuration: BackendConfiguration, provider: Provider = None, **fields
    ):
        super().__init__(configuration, provider, **fields)

    backend_name = None

    def _prepare_job_metadata(self, circuit):
        """Returns the metadata relative to the given circuit that will be attached to the Job"""
        return {
            "qiskit": True,
            "name": circuit.name,
            "num_qubits": circuit.num_qubits,
            "metadata": json.dumps(circuit.metadata),
        }

    @abstractmethod
    def _translate_input(self, circuit):
        pass

    def run(
            self, 
            run_input: Union[QuantumCircuit, List[QuantumCircuit]] = [],
            shots: int = None,
            **options,
        ):
        """Submits the given circuit to run on an Azure Quantum backend.""" 
        circuit = self._normalize_run_input_params(run_input, **options)
        options.pop("run_input", None)
        options.pop("circuit", None)

        # Some Qiskit features require passing lists of circuits, so unpack those here.
        # We currently only support single-experiment jobs.
        if isinstance(circuit, (list, tuple)):
            if len(circuit) > 1:
                raise NotImplementedError("Multi-experiment jobs are not supported!")
            circuit = circuit[0]

        # If the circuit was created using qiskit.assemble,
        # disassemble into QASM here
        if isinstance(circuit, QasmQobj) or isinstance(circuit, PulseQobj):
            from qiskit.assembler import disassemble

            circuits, run, _ = disassemble(circuit)
            circuit = circuits[0]
            if options.get("shots") is None:
                # Note that qiskit.assembler.disassemble() sets the default number of shots for QasmQobj and PulseQobj to 1024
                # unless the user specifies the backend.
                options["shots"] = run["shots"]

        # If not provided as options, the values of these parameters
        # are calculated from the circuit itself:
        job_name = options.pop("job_name", circuit.name)
        metadata = options.pop("metadata", self._prepare_job_metadata(circuit))

        input_params = self._get_input_params(options, shots=shots)

        input_data = self._translate_input(circuit)

        job = super()._run(job_name, input_data, input_params, metadata, **options)

        shots_count = None
        if self._can_send_shots_input_param():
            shots_count = input_params.get(self.__class__._SHOTS_PARAM_NAME)

        logger.info(
            f"Submitted job with id '{job.id()}' for circuit '{circuit.name}' with shot count of {shots_count}:"
        )
        logger.info(input_data)

        return job

def _get_shots_or_deprecated_count_input_param(
        param_name: str,
        shots: int = None, 
        count: int = None,
    ) -> Optional[int]:
    """
    This helper function checks if the deprecated 'count' option is specified.
    In earlier versions it was possible to pass this option to specify shots number for a job,
    but now we only check for it for compatibility reasons.  
    """

    final_shots = None

    if shots is not None:
        final_shots = shots
    
    elif count is not None:
        final_shots = count
        warnings.warn(
            "The 'count' parameter will be deprecated. "
            f"Please, use '{param_name}' parameter instead.",
            category=DeprecationWarning,
        )
    
    return final_shots


##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##
"""
Module providing the Workspace class, used to connect to
an Azure Quantum Workspace.
"""

from __future__ import annotations
from datetime import datetime
import logging
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Optional,
    TYPE_CHECKING,
    Tuple,
    Union,
)
from azure.quantum._client import QuantumClient
from azure.quantum._client.operations import (
    JobsOperations,
    StorageOperations,
    QuotasOperations,
    SessionsOperations,
    TopLevelItemsOperations
)
from azure.quantum._client.models import (
    BlobDetails,
    JobStatus,
    TargetStatus,
)
from azure.quantum import Job, Session
from azure.quantum.job.workspace_item_factory import WorkspaceItemFactory
from azure.quantum._workspace_connection_params import (
    WorkspaceConnectionParams
)
from azure.quantum._constants import (
    ConnectionConstants,
)
from azure.quantum.storage import (
    create_container_using_client,
    get_container_uri,
    ContainerClient
)
if TYPE_CHECKING:
    from azure.quantum.target import Target


logger = logging.getLogger(__name__)

__all__ = ["Workspace"]

# pylint: disable=line-too-long
# pylint: disable=too-many-public-methods
class Workspace:
    """
    Represents an Azure Quantum workspace.

    When creating a Workspace object, callers have two options for
    identifying the Azure Quantum workspace (in order of precedence):
    1. specify a valid location and resource ID; or
    2. specify a valid location, subscription ID, resource group, and workspace name.

    You can also use a connection string to specify the connection parameters
    to an Azure Quantum Workspace by calling
    :obj:`~ Workspace.from_connection_string() <Workspace.from_connection_string>`.

    If the Azure Quantum workspace does not have linked storage, the caller
    must also pass a valid Azure storage account connection string.

    :param subscription_id:
        The Azure subscription ID.
        Ignored if resource_id is specified.

    :param resource_group:
        The Azure resource group name.
        Ignored if resource_id is specified.

    :param name:
        The Azure Quantum workspace name.
        Ignored if resource_id is specified.

    :param storage:
        The Azure storage account connection string.
        Required only if the specified Azure Quantum
        workspace does not have linked storage.

    :param resource_id:
        The resource ID of the Azure Quantum workspace.

    :param location:
        The Azure region where the Azure Quantum workspace is provisioned.
        This may be specified as a region name such as
        \"East US\" or a location name such as \"eastus\".

    :param credential:
        The credential to use to connect to Azure services.
        Normally one of the credential types from [Azure.Identity](https://learn.microsoft.com/python/api/overview/azure/identity-readme?view=azure-python#credential-classes).

        Defaults to \"DefaultAzureCredential\", which will attempt multiple
        forms of authentication.

    :param user_agent:
        Add the specified value as a prefix to the HTTP User-Agent header
        when communicating to the Azure Quantum service.
    """
    def __init__(
        self,
        subscription_id: Optional[str] = None,
        resource_group: Optional[str] = None,
        name: Optional[str] = None,
        storage: Optional[str] = None,
        resource_id: Optional[str] = None,
        location: Optional[str] = None,
        credential: Optional[object] = None,
        user_agent: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        connection_params = WorkspaceConnectionParams(
            location=location,
            subscription_id=subscription_id,
            resource_group=resource_group,
            workspace_name=name,
            credential=credential,
            resource_id=resource_id,
            user_agent=user_agent,
            **kwargs
        ).default_from_env_vars()

        logger.info("Using %s environment.", connection_params.environment)

        connection_params.assert_complete()

        connection_params.on_new_client_request = self._on_new_client_request

        self._connection_params = connection_params
        self._storage = storage

        # Create QuantumClient
        self._client = self._create_client()

    def _on_new_client_request(self) -> None:
        """
        An internal callback method used by the WorkspaceConnectionParams
        to ask the Workspace to recreate the underlying Azure SDK REST API client.
        This is used when some value (such as the UserAgent) has changed
        in the WorkspaceConnectionParams and requires the client to be
        recreated.
        """
        self._client = self._create_client()

    @property
    def location(self) -> str:
        """
        Returns the Azure location of the Quantum Workspace.

        :return: Azure location name.
        :rtype: str
        """
        return self._connection_params.location

    @property
    def subscription_id(self) -> str:
        """
        Returns the Azure Subscription ID of the Quantum Workspace.

        :return: Azure Subscription ID.
        :rtype: str
        """
        return self._connection_params.subscription_id

    @property
    def resource_group(self) -> str:
        """
        Returns the Azure Resource Group of the Quantum Workspace.

        :return: Azure Resource Group name.
        :rtype: str
        """
        return self._connection_params.resource_group

    @property
    def name(self) -> str:
        """
        Returns the Name of the Quantum Workspace.

        :return: Azure Quantum Workspace name.
        :rtype: str
        """
        return self._connection_params.workspace_name

    @property
    def credential(self) -> Any:
        """
        Returns the Credential used to connect to the Quantum Workspace.

        :return: Azure SDK Credential from [Azure.Identity](https://learn.microsoft.com/python/api/overview/azure/identity-readme?view=azure-python#credential-classes).
        :rtype: typing.Any
        """
        return self._connection_params.credential

    @property
    def storage(self) -> str:
        """
        Returns the Azure Storage account name associated with the Quantum Workspace.

        :return: Azure Storage account name.
        :rtype: str
        """
        return self._storage

    def _create_client(self) -> QuantumClient:
        """"
        An internal method to (re)create the underlying Azure SDK REST API client.

        :return: Azure SDK REST API client for Azure Quantum.
        :rtype: QuantumClient
        """
        connection_params = self._connection_params
        kwargs = {}
        if connection_params.api_version:
            kwargs["api_version"] = connection_params.api_version
        client = QuantumClient(
            credential=connection_params.get_credential_or_default(),
            subscription_id=connection_params.subscription_id,
            resource_group_name=connection_params.resource_group,
            workspace_name=connection_params.workspace_name,
            azure_region=connection_params.location,
            user_agent=connection_params.get_full_user_agent(),
            credential_scopes = [ConnectionConstants.DATA_PLANE_CREDENTIAL_SCOPE],
            endpoint=connection_params.quantum_endpoint,
            authentication_policy=connection_params.get_auth_policy(),
            **kwargs
        )
        return client

    @property
    def user_agent(self) -> str:
        """
        Returns the Workspace's UserAgent string that is sent to
        the service via the UserAgent header.

        :return: User Agent string.
        :rtype: str
        """
        return self._connection_params.get_full_user_agent()

    def append_user_agent(self, value: str) -> None:
        """
        Append a new value to the Workspace's UserAgent.
        The values are appended using a dash.

        :param value:
            UserAgent value to add, e.g. "azure-quantum-<plugin>"
        """
        self._connection_params.append_user_agent(value=value)

    @classmethod
    def from_connection_string(cls, connection_string: str, **kwargs) -> Workspace:
        """
        Creates a new Azure Quantum Workspace client from a connection string.

        :param connection_string:
            A valid connection string, usually obtained from the
            `Quantum Workspace -> Operations -> Access Keys` blade in the Azure Portal.

        :return: New Azure Quantum Workspace client.
        :rtype: Workspace
        """
        connection_params = WorkspaceConnectionParams(connection_string=connection_string)
        return cls(
            subscription_id=connection_params.subscription_id,
            resource_group=connection_params.resource_group,
            name=connection_params.workspace_name,
            location=connection_params.location,
            credential=connection_params.get_credential_or_default(),
            **kwargs)

    def _get_top_level_items_client(self) -> TopLevelItemsOperations:
        """
        Returns the internal Azure SDK REST API client
        for the `{workspace}/topLevelItems` API.

        :return: REST API client for the `topLevelItems` API.
        :rtype: TopLevelItemsOperations
        """
        return self._client.top_level_items

    def _get_sessions_client(self) -> SessionsOperations:
        """
        Returns the internal Azure SDK REST API client
        for the `{workspace}/sessions` API.

        :return: REST API client for the `sessions` API.
        :rtype: SessionsOperations
        """
        return self._client.sessions

    def _get_jobs_client(self) -> JobsOperations:
        """
        Returns the internal Azure SDK REST API client
        for the `{workspace}/jobs` API.

        :return: REST API client for the `jobs` API.
        :rtype: JobsOperations
        """
        return self._client.jobs

    def _get_workspace_storage_client(self) -> StorageOperations:
        """
        Returns the internal Azure SDK REST API client
        for the `{workspace}/storage` API.

        :return: REST API client for the `storage` API.
        :rtype: StorageOperations
        """
        return self._client.storage

    def _get_quotas_client(self) -> QuotasOperations:
        """
        Returns the internal Azure SDK REST API client
        for the `{workspace}/quotas` API.

        :return: REST API client for the `quotas` API.
        :rtype: QuotasOperations
        """
        return self._client.quotas

    def _get_linked_storage_sas_uri(
        self,
        container_name: str,
        blob_name: Optional[str] = None
    ) -> str:
        """
        Calls the service and returns a container/blob SAS URL
        for the Storage associated with the Quantum Workspace.

        :param container_name:
            The name of the storage container.

        :param blob_name:
            Optional name of the blob. Defaults to `None`.

        :return: Storage Account SAS URL to a container or blob.
        :rtype: str
        """
        client = self._get_workspace_storage_client()
        blob_details = BlobDetails(
            container_name=container_name, blob_name=blob_name
        )
        container_uri = client.sas_uri(blob_details=blob_details)

        logger.debug("Container URI from service: %s", container_uri)
        return container_uri.sas_uri

    def submit_job(self, job: Job) -> Job:
        """
        Submits a job to be processed in the Workspace.

        :param job:
            Job to submit.

        :return: Azure Quantum Job that was submitted, with an updated status.
        :rtype: Job
        """
        client = self._get_jobs_client()
        details = client.create(
            job.details.id, job.details
        )
        return Job(self, details)

    def cancel_job(self, job: Job) -> Job:
        """
        Requests the Workspace to cancel the
        execution of a job.

        :param job:
            Job to cancel.

        :return: Azure Quantum Job that was requested to be cancelled, with an updated status.
        :rtype: Job
        """
        client = self._get_jobs_client()
        client.cancel(job.details.id)
        details = client.get(job.id)
        return Job(self, details)

    def get_job(self, job_id: str) -> Job:
        """
        Returns the job corresponding to the given id.
        
        :param job_id:
            Id of a job to fetch.

        :return: Azure Quantum Job.
        :rtype: Job
        """
        # pylint: disable=import-outside-toplevel
        from azure.quantum.target.target_factory import TargetFactory
        from azure.quantum.target import Target

        client = self._get_jobs_client()
        details = client.get(job_id)
        target_factory = TargetFactory(base_cls=Target, workspace=self)
        # pylint: disable=protected-access
        target_cls = target_factory._target_cls(
            details.provider_id,
            details.target)
        job_cls = target_cls._get_job_class()
        return job_cls(self, details)

    def list_jobs(
        self,
        name_match: Optional[str] = None,
        status: Optional[JobStatus] = None,
        created_after: Optional[datetime] = None
    ) -> List[Job]:
        """
        Returns list of jobs that meet optional (limited) filter criteria.

        :param name_match:
            Optional Regular Expression for job name matching. Defaults to `None`.

        :param status:
            Optional filter by job status. Defaults to `None`.

        :param created_after:
            Optional filter by jobs that were created after the given time. Defaults to `None`.

        :return: Jobs that matched the search criteria.
        :rtype: typing.List[Job]
        """
        client = self._get_jobs_client()
        jobs = client.list()

        result = []
        for j in jobs:
            deserialized_job = Job(self, j)
            if deserialized_job.matches_filter(name_match, status, created_after):
                result.append(deserialized_job)

        return result

    def _get_target_status(
            self,
            name: Optional[str] = None,
            provider_id: Optional[str] = None,
        ) -> List[Tuple[str, TargetStatus]]:
        """
        Returns a list of tuples containing the `Provider ID` and `Target Status`,
        with the option of filtering that list by a combination of Provider ID and Target Name.

        :param name:
            Optional name of the Target to filter for. Defaults to `None`.

        :param provider_id:
            Optional Provider ID to filter for. Defaults to `None`.

        :return: List of tuples containing Provider ID and TargetStatus.
        :rtype: typing.List[typing.Tuple[str, TargetStatus]]
        """
        return [
            (provider.id, target)
            for provider in self._client.providers.get_status()
            for target in provider.targets
            if (provider_id is None or provider.id.lower() == provider_id.lower())
                and (name is None or target.id.lower() == name.lower())
        ]

    def get_targets(
        self,
        name: Optional[str] = None,
        provider_id: Optional[str] = None,
    ) -> Union[Target, Iterable[Target]]:
        """
        Returns all available targets for this workspace filtered by Target name and Provider ID.
        If the target name is passed, a single `Target` object will be returned.
        Otherwise it returns a iterable/list of `Target` objects, optionally filtered by the Provider ID.

        :param name:
            Optional target name to filter by, defaults to `None`.
        
        :param provider_id:
            Optional provider Id to filter by, defaults to `None`.

        :return: A single Azure Quantum Target or a iterable/list of Targets.
        :rtype: typing.Union[Target, typing.Iterable[Target]]
        """
        # pylint: disable=import-outside-toplevel
        from azure.quantum.target.target_factory import TargetFactory
        from azure.quantum.target import Target

        target_factory = TargetFactory(
            base_cls=Target,
            workspace=self
        )
        return target_factory.get_targets(
            name=name,
            provider_id=provider_id
        )

    def get_quotas(self) -> List[Dict[str, Any]]:
        """
        Get a list of quotas for the given workspace.
        Each quota is represented as a dictionary, containing the
        properties for that quota.

        Common Quota properties are:
        - "dimension": The dimension that the quota is applied to. 
        - "scope": The scope that the quota is applied to.
        - "provider_id": The provider that the quota is applied to.
        - "utilization": The current utilization of the quota.
        - "limit": The limit of the quota.
        - "period": The period that the quota is applied to.

        :return: Workspace quotas.
        :rtype: typing.List[typing.Dict[str, typing.Any]
        """
        client = self._get_quotas_client()
        return [q.as_dict() for q in client.list()]

    def list_top_level_items(
        self
    ) -> List[Union[Job, Session]]:
        """
        Get a list of top level items for the given workspace,
        which can be standalone Jobs (Jobs not associated with a Session)
        or Sessions (which can contain Jobs).

        :return: List of Workspace top level Jobs or Sessions.
        :rtype: typing.List[typing.Union[Job, Session]]
        """
        client = self._get_top_level_items_client()
        item_details_list = client.list()
        result = [WorkspaceItemFactory.__new__(workspace=self, item_details=item_details)
                  for item_details in item_details_list]
        return result

    def list_sessions(
        self
    ) -> List[Session]:
        """
        Get the list of sessions in the given workspace.

        :return: List of Workspace Sessions.
        :rtype: typing.List[Session]
        """
        client = self._get_sessions_client()
        session_details_list = client.list()
        result = [Session(workspace=self,details=session_details)
                  for session_details in session_details_list]
        return result

    def open_session(
        self,
        session: Session,
    ) -> None:
        """
        Opens/creates a session in the given workspace.

        :param session:
            The session to be opened/created.

        :return: A new open Azure Quantum Session.
        :rtype: Session
        """
        client = self._get_sessions_client()
        session.details = client.open(
            session_id=session.id,
            session=session.details)

    def close_session(
        self,
        session: Session
    ) -> None:
        """
        Closes a session in the given workspace if the
        session is not in a terminal state.
        Otherwise, just refreshes the session details.

        :param session:
            The session to be closed.
        """
        client = self._get_sessions_client()
        if not session.is_in_terminal_state():
            session.details = client.close(session_id=session.id)
        else:
            session.details = client.get(session_id=session.id)

        if session.target:
            if (session.target.latest_session
                and session.target.latest_session.id == session.id):
                session.target.latest_session.details = session.details

    def refresh_session(
        self,
        session: Session
    ) -> None:
        """
        Updates the session details with the latest information
        from the workspace.

        :param session:
            The session to be refreshed.
        """
        session.details = self.get_session(session_id=session.id).details

    def get_session(
        self,
        session_id: str
    ) -> Session:
        """
        Gets a session from the workspace.

        :param session_id:
            The id of session to be retrieved.

        :return: Azure Quantum Session
        :rtype: Session
        """
        client = self._get_sessions_client()
        session_details = client.get(session_id=session_id)
        result = Session(workspace=self, details=session_details)
        return result

    def list_session_jobs(
        self,
        session_id: str
    ) -> List[Job]:
        """
        Gets all jobs associated with a session.

        :param session_id:
            The id of session.

        :return: List of all jobs associated with a session.
        :rtype: typing.List[Job]
        """
        client = self._get_sessions_client()
        job_details_list = client.jobs_list(session_id=session_id)
        result = [Job(workspace=self, job_details=job_details)
                  for job_details in job_details_list]
        return result

    def get_container_uri(
        self,
        job_id: Optional[str] = None,
        container_name: Optional[str] = None,
        container_name_format: Optional[str] = "job-{job_id}"
    ) -> str:
        """
        Get container URI based on job ID or container name.
        Creates a new container if it does not yet exist.

        :param job_id:
            Job ID, defaults to `None`.

        :param container_name:
            Container name, defaults to `None`.

        :param container_name_format:
            Container name format, defaults to "job-{job_id}".
        
        :return: Container URI.
        :rtype: str
        """
        if container_name is None:
            if job_id is not None:
                container_name = container_name_format.format(job_id=job_id)
            elif job_id is None:
                container_name = f"{self.name}-data"
        # Create container URI and get container client
        if self.storage is None:
            # Get linked storage account from the service, create
            # a new container if it does not yet exist
            container_uri = self._get_linked_storage_sas_uri(
                container_name
            )
            container_client = ContainerClient.from_container_url(
                container_uri
            )
            create_container_using_client(container_client)
        else:
            # Use the storage acount specified to generate container URI,
            # create a new container if it does not yet exist
            container_uri = get_container_uri(
                self.storage, container_name
            )
        return container_uri



# ------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# ------------------------------------
import logging
import re
from typing import Optional
import urllib3
from azure.core.credentials import AccessToken
from azure.identity import (
    AzurePowerShellCredential,
    EnvironmentCredential,
    ManagedIdentityCredential,
    AzureCliCredential,
    VisualStudioCodeCredential,
    InteractiveBrowserCredential,
    DeviceCodeCredential,
    _internal as AzureIdentityInternals,
)
from ._chained import _ChainedTokenCredential
from ._token import _TokenFileCredential
from azure.quantum._constants import ConnectionConstants

_LOGGER = logging.getLogger(__name__)
WWW_AUTHENTICATE_REGEX = re.compile(
    r"""
        ^
        Bearer\sauthorization_uri="
            https://(?P<authority>[^/]*)/
            (?P<tenant_id>[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12})
        "
    """,
    re.VERBOSE | re.IGNORECASE)
WWW_AUTHENTICATE_HEADER_NAME = "WWW-Authenticate"


class _DefaultAzureCredential(_ChainedTokenCredential):
    """
    Based on Azure.Identity.DefaultAzureCredential from:
    https://github.com/Azure/azure-sdk-for-python/blob/master/sdk/identity/azure-identity/azure/identity/_credentials/default.py

    The three key differences are:
    1) Inherit from _ChainedTokenCredential, which has 
       more aggressive error handling than ChainedTokenCredential
    2) Instantiate the internal credentials the first time the get_token gets called
       such that we can get the tenant_id if it was not passed by the user (but we don't
       want to do that in the constructor).
       We automatically identify the user's tenant_id for a given subscription 
       so that users with MSA accounts don't need to pass it.
       This is a mitigation for bug https://github.com/Azure/azure-sdk-for-python/issues/18975
       We need the following parameters to enable auto-detection of tenant_id
       - subscription_id
       - arm_endpoint (defaults to the production url "https://management.azure.com/")
    3) Add custom TokenFileCredential as first method to attempt,
       which will look for a local access token.
    """
    def __init__(
        self,
        arm_endpoint: str,
        subscription_id: str,
        client_id: Optional[str] = None,
        tenant_id: Optional[str] = None,
        authority: Optional[str] = None,
    ):
        if arm_endpoint is None:
            raise ValueError("arm_endpoint is mandatory parameter")
        if subscription_id is None:
            raise ValueError("subscription_id is mandatory parameter")

        self.authority = self._authority_or_default(
            authority=authority,
            arm_endpoint=arm_endpoint)
        self.tenant_id = tenant_id
        self.subscription_id = subscription_id
        self.arm_endpoint = arm_endpoint
        self.client_id = client_id
        # credentials will be created lazy on the first call to get_token
        super(_DefaultAzureCredential, self).__init__()

    def _authority_or_default(self, authority: str, arm_endpoint: str):
        if authority:
            return AzureIdentityInternals.normalize_authority(authority)
        if arm_endpoint == ConnectionConstants.ARM_DOGFOOD_ENDPOINT:
            return ConnectionConstants.DOGFOOD_AUTHORITY
        return ConnectionConstants.AUTHORITY

    def _initialize_credentials(self):
        self._discover_tenant_id_(
            arm_endpoint=self.arm_endpoint,
            subscription_id=self.subscription_id)
        credentials = []
        credentials.append(_TokenFileCredential())
        credentials.append(EnvironmentCredential())
        if self.client_id:
            credentials.append(ManagedIdentityCredential(client_id=self.client_id))
        if self.authority and self.tenant_id:
            credentials.append(VisualStudioCodeCredential(authority=self.authority, tenant_id=self.tenant_id))
            credentials.append(AzureCliCredential(tenant_id=self.tenant_id))
            credentials.append(AzurePowerShellCredential(tenant_id=self.tenant_id))
            credentials.append(InteractiveBrowserCredential(authority=self.authority, tenant_id=self.tenant_id))
            if self.client_id:
                credentials.append(DeviceCodeCredential(authority=self.authority, client_id=self.client_id, tenant_id=self.tenant_id))
        self.credentials = credentials

    def get_token(self, *scopes: str, **kwargs) -> AccessToken:
        """
        Request an access token for `scopes`.
        This method is called automatically by Azure SDK clients.
        
        :param str scopes: desired scopes for the access token.
        This method requires at least one scope.
        
        :raises ~azure.core.exceptions.ClientAuthenticationError:authentication failed.
            The exception has a `message` attribute listing each authentication
            attempt and its error message.
        """
        # lazy-initialize the credentials
        if self.credentials is None or len(self.credentials) == 0:
            self._initialize_credentials()

        return super(_DefaultAzureCredential, self).get_token(*scopes, **kwargs)

    def _discover_tenant_id_(self, arm_endpoint:str, subscription_id:str):
        """
        If the tenant_id was not given, try to obtain it
        by calling the management endpoint for the subscription_id,
        or by applying default values.
        """
        if self.tenant_id:
            return

        try:
            url = (
                f"{arm_endpoint.rstrip('/')}/subscriptions/"
                + f"{subscription_id}?api-version=2018-01-01"
                + "&discover-tenant-id"  # used by the test recording infrastructure
            )
            http = urllib3.PoolManager()
            response = http.request(
                method="GET",
                url=url,
            )
            if WWW_AUTHENTICATE_HEADER_NAME in response.headers:
                www_authenticate = response.headers[WWW_AUTHENTICATE_HEADER_NAME]
                match = re.search(WWW_AUTHENTICATE_REGEX, www_authenticate)
                if match:
                    self.tenant_id = match.group("tenant_id")
        # pylint: disable=broad-exception-caught
        except Exception as ex:
            _LOGGER.error(ex)

        # apply default values
        self.tenant_id = self.tenant_id or ConnectionConstants.MSA_TENANT_ID



##
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
##

from typing import TYPE_CHECKING, Any, Dict, List, Union
from azure.quantum.version import __version__
from qiskit import QuantumCircuit
from abc import abstractmethod
from .backend import AzureQirBackend, QIR_BASIS_GATES

from qiskit.providers.models import BackendConfiguration
from qiskit.providers import Options, Provider
from qsharp import TargetProfile
from qsharp.interop.qiskit import ResourceEstimatorBackend
import pyqir as pyqir

if TYPE_CHECKING:
    from azure.quantum.qiskit import AzureQuantumProvider

import logging

logger = logging.getLogger(__name__)

__all__ = ["MicrosoftBackend", "MicrosoftResourceEstimationBackend"]


class MicrosoftBackend(AzureQirBackend):
    """Base class for interfacing with a Microsoft backend in Azure Quantum"""

    @abstractmethod
    def __init__(
        self, configuration: BackendConfiguration, provider: Provider = None, **fields
    ):
        super().__init__(configuration, provider, **fields)

    @classmethod
    def _default_options(cls):
        return Options(target_profile=TargetProfile.Adaptive_RI)

    def _azure_config(self) -> Dict[str, str]:
        config = super()._azure_config()
        config.update(
            {
                "provider_id": "microsoft-qc",
                "output_data_format": "microsoft.resource-estimates.v1",
            }
        )
        return config

    def _translate_input(
        self,
        circuits: Union[QuantumCircuit, List[QuantumCircuit]],
        input_params: Dict[str, Any],
    ) -> bytes:
        """Translates the input values to the QIR expected by the Backend."""
        # All the logic is in the base class, but we need to override
        # this method to ensure that the bitcode QIR format is used for RE.

        # normal translation is to QIR text format in utf-8 encoded bytes
        ir_byte_str = super()._translate_input(circuits, input_params)
        # decode the utf-8 encoded bytes to a string
        ir_str = ir_byte_str.decode('utf-8')
        # convert the QIR text format to QIR bitcode format
        module = pyqir.Module.from_ir(pyqir.Context(), ir_str)
        return module.bitcode

    def _generate_qir(
        self, circuits: List[QuantumCircuit], target_profile: TargetProfile, **kwargs
    ) -> pyqir.Module:
        if len(circuits) == 0:
            raise ValueError("No QuantumCircuits provided")

        name = "circuits"
        if isinstance(circuits, QuantumCircuit):
            name = circuits.name
            circuits = [circuits]
        elif isinstance(circuits, list):
            for value in circuits:
                if not isinstance(value, QuantumCircuit):
                    raise ValueError(
                        "Input must be Union[QuantumCircuit, List[QuantumCircuit]]"
                    )
        else:
            raise ValueError(
                "Input must be Union[QuantumCircuit, List[QuantumCircuit]]"
            )

        skip_transpilation = kwargs.pop("skip_transpilation", False)
        backend = ResourceEstimatorBackend(
            skip_transpilation=skip_transpilation, **kwargs
        )
        context = pyqir.Context()
        llvm_module = pyqir.qir_module(context, name)
        for circuit in circuits:
            qir_str = backend.qir(circuit, target_profile=target_profile)
            module = pyqir.Module.from_ir(context, qir_str)
            llvm_module.link(module)

        err = llvm_module.verify()
        if err is not None:
            raise Exception(err)

        return llvm_module


class MicrosoftResourceEstimationBackend(MicrosoftBackend):
    """Backend class for interfacing with the resource estimator target"""

    backend_names = ("microsoft.estimator",)

    @classmethod
    def _default_options(cls):
        return Options(
            target_profile=TargetProfile.Adaptive_RI,
            errorBudget=1e-3,
            qubitParams={"name": "qubit_gate_ns_e3"},
            qecScheme={"name": "surface_code"}
        )

    def __init__(self, name: str, provider: "AzureQuantumProvider", **kwargs):
        """Constructor for class to interface with the resource estimator target"""
        default_config = BackendConfiguration.from_dict(
            {
                "backend_name": name,
                "backend_version": __version__,
                "simulator": True,
                "local": False,
                "coupling_map": None,
                "description": "Resource estimator on Azure Quantum",
                "basis_gates": QIR_BASIS_GATES,
                "memory": False,
                "n_qubits": 0xFFFFFFFFFFFFFFFF,  # NOTE: maximum 64-bit unsigned value
                "conditional": True,
                "max_shots": 1,
                "max_experiments": 1,
                "open_pulse": False,
                "gates": [
                    {"name": "TODO", "parameters": [], "qasm_def": "TODO"}
                ],  # NOTE: copied from other backends
                "azure": self._azure_config(),
                "is_default": True,
            }
        )
        logger.info("Initializing MicrosoftResourceEstimationBackend")
        configuration: BackendConfiguration = kwargs.pop(
            "configuration", default_config
        )
        super().__init__(configuration=configuration, provider=provider, **kwargs)



##
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
##

from typing import TYPE_CHECKING, Dict
from azure.quantum.version import __version__
from azure.quantum.target.rigetti import RigettiTarget
from abc import abstractmethod
from .backend import AzureQirBackend

from qiskit.providers.models import BackendConfiguration
from qiskit.providers import Options, Provider
from qsharp import TargetProfile

if TYPE_CHECKING:
    from azure.quantum.qiskit import AzureQuantumProvider

import logging

logger = logging.getLogger(__name__)

__all__ = ["RigettiSimulatorBackend" "RigettiQPUBackend"]


_DEFAULT_SHOTS_COUNT = 500

class RigettiBackend(AzureQirBackend):
    """Base class for interfacing with a Rigetti backend in Azure Quantum"""

    _SHOTS_PARAM_NAME = "count"

    @abstractmethod
    def __init__(
        self, configuration: BackendConfiguration, provider: Provider = None, **fields
    ):
        super().__init__(configuration, provider, **fields)

    @classmethod
    def _default_options(cls):
        other_options = {
            cls._SHOTS_PARAM_NAME: _DEFAULT_SHOTS_COUNT,
        }
        return Options(target_profile=TargetProfile.Base, **other_options)

    def _azure_config(self) -> Dict[str, str]:
        config = super()._azure_config()
        config.update(
            {
                "provider_id": "rigetti",
            }
        )
        return config


class RigettiSimulatorBackend(RigettiBackend):
    backend_names = RigettiTarget.simulators()

    def __init__(self, name: str, provider: "AzureQuantumProvider", **kwargs):
        """Base class for interfacing with an Rigetti Simulator backend"""
        default_config = BackendConfiguration.from_dict(
            {
                "backend_name": name,
                "backend_version": __version__,
                "simulator": True,
                "local": False,
                "coupling_map": None,
                "description": "Rigetti simulator on Azure Quantum",
                "basis_gates": self._basis_gates(),
                "memory": True,
                "n_qubits": RigettiTarget.num_qubits(name),
                "conditional": False,
                "max_shots": 10000,
                "max_experiments": 1,
                "open_pulse": False,
                "gates": [{"name": "TODO", "parameters": [], "qasm_def": "TODO"}],
                "azure": self._azure_config(),
                "is_default": True,
            }
        )
        logger.info("Initializing RigettiSimulatorBackend")
        configuration: BackendConfiguration = kwargs.pop(
            "configuration", default_config
        )
        super().__init__(configuration=configuration, provider=provider, **kwargs)


class RigettiQPUBackend(RigettiBackend):
    backend_names = RigettiTarget.qpus()

    def __init__(self, name: str, provider: "AzureQuantumProvider", **kwargs):
        """Base class for interfacing with a Rigetti QPU backend"""
        default_config = BackendConfiguration.from_dict(
            {
                "backend_name": name,
                "backend_version": __version__,
                "simulator": False,
                "local": False,
                "coupling_map": None,
                "description": "Rigetti QPU on Azure Quantum",
                "basis_gates": self._basis_gates(),
                "memory": True,
                "n_qubits": RigettiTarget.num_qubits(name),
                "conditional": False,
                "max_shots": 10000,
                "max_experiments": 1,
                "open_pulse": False,
                "gates": [{"name": "TODO", "parameters": [], "qasm_def": "TODO"}],
                "azure": self._azure_config(),
                "is_default": True,
            }
        )
        logger.info("Initializing RigettiQPUBackend")
        configuration: BackendConfiguration = kwargs.pop(
            "configuration", default_config
        )
        super().__init__(configuration=configuration, provider=provider, **kwargs)



##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##
from __future__ import annotations
import re
import os
from re import Match
from typing import (
    Optional,
    Callable,
    Union,
    Any
)
from azure.core.credentials import AzureKeyCredential
from azure.core.pipeline.policies import AzureKeyCredentialPolicy
from azure.quantum._authentication import _DefaultAzureCredential
from azure.quantum._constants import (
    EnvironmentKind,
    EnvironmentVariables,
    ConnectionConstants,
    GUID_REGEX_PATTERN,
)

class WorkspaceConnectionParams:
    """
    Internal Azure Quantum Python SDK class to handle logic
    for the parameters needed to connect to a Workspace.
    """

    RESOURCE_ID_REGEX = re.compile(
        fr"""
            ^
            /subscriptions/(?P<subscription_id>{GUID_REGEX_PATTERN})
            /resourceGroups/(?P<resource_group>[^\s/]+)
            /providers/Microsoft\.Quantum
            /Workspaces/(?P<workspace_name>[^\s/]+)
            $
        """,
        re.VERBOSE | re.IGNORECASE)

    CONNECTION_STRING_REGEX = re.compile(
        fr"""
            ^
            SubscriptionId=(?P<subscription_id>{GUID_REGEX_PATTERN});
            ResourceGroupName=(?P<resource_group>[^\s;]+);
            WorkspaceName=(?P<workspace_name>[^\s;]+);
            ApiKey=(?P<api_key>[^\s;]+);
            QuantumEndpoint=(?P<quantum_endpoint>https://(?P<location>[^\s\.]+).quantum(?:-test)?.azure.com/);
        """,
        re.VERBOSE | re.IGNORECASE)

    def __init__(
        self,
        subscription_id: Optional[str] = None,
        resource_group: Optional[str] = None,
        workspace_name: Optional[str] = None,
        location: Optional[str] = None,
        quantum_endpoint: Optional[str] = None,
        arm_endpoint: Optional[str] = None,
        environment: Union[str, EnvironmentKind, None] = None,
        credential: Optional[object] = None,
        resource_id: Optional[str] = None,
        user_agent: Optional[str] = None,
        user_agent_app_id: Optional[str] = None,
        tenant_id: Optional[str] = None,
        client_id: Optional[str] = None,
        api_version: Optional[str] = None,
        connection_string: Optional[str] = None,
        on_new_client_request: Optional[Callable] = None,
    ):
        # fields are used for these properties since
        # they have special getters/setters
        self._location = None
        self._environment = None
        self._quantum_endpoint = None
        self._arm_endpoint = None
        # regular connection properties
        self.subscription_id = None
        self.resource_group = None
        self.workspace_name = None
        self.credential = None
        self.user_agent = None
        self.user_agent_app_id = None
        self.client_id = None
        self.tenant_id = None
        self.api_version = None
        # callback to create a new client if needed
        # for example, when changing the user agent
        self.on_new_client_request = on_new_client_request
        # merge the connection parameters passed
        # connection_string is set first as it
        # should be overridden by other parameters
        self.apply_connection_string(connection_string)
        self.merge(
            api_version=api_version,
            arm_endpoint=arm_endpoint,
            quantum_endpoint=quantum_endpoint,
            client_id=client_id,
            credential=credential,
            environment=environment,
            location=location,
            resource_group=resource_group,
            subscription_id=subscription_id,
            tenant_id=tenant_id,
            user_agent=user_agent,
            user_agent_app_id=user_agent_app_id,
            workspace_name=workspace_name,
        )
        self.apply_resource_id(resource_id=resource_id)

    @property
    def location(self):
        """
        The Azure location.
        On the setter, we normalize the value removing spaces
        and converting it to lowercase.
        """
        return self._location

    @location.setter
    def location(self, value: str):
        self._location = (value.replace(" ", "").lower()
                          if isinstance(value, str)
                          else value)

    @property
    def environment(self):
        """
        The environment kind, such as dogfood, canary or production.
        Defaults to EnvironmentKind.PRODUCTION
        """
        return self._environment or EnvironmentKind.PRODUCTION

    @environment.setter
    def environment(self, value: Union[str, EnvironmentKind]):
        self._environment = (EnvironmentKind[value.upper()]
                             if isinstance(value, str)
                             else value)

    @property
    def quantum_endpoint(self):
        """
        The Azure Quantum data plane endpoint.
        Defaults to well-known endpoint based on the environment.
        """
        if self._quantum_endpoint:
            return self._quantum_endpoint
        if not self.location:
            raise ValueError("Location not specified")
        if self.environment is EnvironmentKind.PRODUCTION:
            return ConnectionConstants.GET_QUANTUM_PRODUCTION_ENDPOINT(self.location)
        if self.environment is EnvironmentKind.CANARY:
            return ConnectionConstants.GET_QUANTUM_CANARY_ENDPOINT(self.location)
        if self.environment is EnvironmentKind.DOGFOOD:
            return ConnectionConstants.GET_QUANTUM_DOGFOOD_ENDPOINT(self.location)
        raise ValueError(f"Unknown environment `{self.environment}`.")

    @quantum_endpoint.setter
    def quantum_endpoint(self, value: str):
        self._quantum_endpoint = value

    @property
    def arm_endpoint(self):
        """
        The control plane endpoint.
        Defaults to well-known arm_endpoint based on the environment.
        """
        if self._arm_endpoint:
            return self._arm_endpoint
        if self.environment is EnvironmentKind.DOGFOOD:
            return ConnectionConstants.ARM_DOGFOOD_ENDPOINT
        if self.environment in [EnvironmentKind.PRODUCTION,
                                EnvironmentKind.CANARY]:
            return ConnectionConstants.ARM_PRODUCTION_ENDPOINT
        raise ValueError(f"Unknown environment `{self.environment}`.")

    @arm_endpoint.setter
    def arm_endpoint(self, value: str):
        self._arm_endpoint = value

    @property
    def api_key(self):
        """
        The api-key stored in a AzureKeyCredential.
        """
        return (self.credential.key
                if isinstance(self.credential, AzureKeyCredential)
                else None)

    @api_key.setter
    def api_key(self, value: str):
        if value:
            self.credential = AzureKeyCredential(value)
        self._api_key = value

    def __repr__(self):
        """
        Print all fields and properties.
        """
        info = []
        for key in vars(self):
            info.append(f"    {key}: {self.__dict__[key]}")
        cls = type(self)
        for key in dir(self):
            attr = getattr(cls, key, None)
            if attr and isinstance(attr, property) and attr.fget:
                info.append(f"    {key}: {attr.fget(self)}")
        info.sort()
        info.insert(0, super().__repr__())
        return "\n".join(info)

    def apply_resource_id(self, resource_id: str):
        """
        Parses the resource_id and set the connection
        parameters obtained from it.
        """
        if resource_id:
            match = re.search(
                WorkspaceConnectionParams.RESOURCE_ID_REGEX,
                resource_id)
            if not match:
                raise ValueError("Invalid resource id")
            self._merge_re_match(match)

    def apply_connection_string(self, connection_string: str):
        """
        Parses the connection_string and set the connection
        parameters obtained from it.
        """
        if connection_string:
            match = re.search(
                WorkspaceConnectionParams.CONNECTION_STRING_REGEX,
                connection_string)
            if not match:
                raise ValueError("Invalid connection string")
            self._merge_re_match(match)

    def merge(
        self,
        subscription_id: Optional[str] = None,
        resource_group: Optional[str] = None,
        workspace_name: Optional[str] = None,
        location: Optional[str] = None,
        quantum_endpoint: Optional[str] = None,
        arm_endpoint: Optional[str] = None,
        environment: Union[str, EnvironmentKind, None] = None,
        credential: Optional[object] = None,
        user_agent: Optional[str] = None,
        user_agent_app_id: Optional[str] = None,
        tenant_id: Optional[str] = None,
        client_id: Optional[str] = None,
        api_version: Optional[str] = None,
        api_key: Optional[str] = None,
    ):
        """
        Set all fields/properties with `not None` values
        passed in the (named or key-valued) arguments
        into this instance.
        """
        self._merge(
            api_version=api_version,
            arm_endpoint=arm_endpoint,
            quantum_endpoint=quantum_endpoint,
            client_id=client_id,
            credential=credential,
            environment=environment,
            location=location,
            resource_group=resource_group,
            subscription_id=subscription_id,
            tenant_id=tenant_id,
            user_agent=user_agent,
            user_agent_app_id=user_agent_app_id,
            workspace_name=workspace_name,
            api_key=api_key,
            merge_default_mode=False,
        )
        return self

    def apply_defaults(
        self,
        subscription_id: Optional[str] = None,
        resource_group: Optional[str] = None,
        workspace_name: Optional[str] = None,
        location: Optional[str] = None,
        quantum_endpoint: Optional[str] = None,
        arm_endpoint: Optional[str] = None,
        environment: Union[str, EnvironmentKind, None] = None,
        credential: Optional[object] = None,
        user_agent: Optional[str] = None,
        user_agent_app_id: Optional[str] = None,
        tenant_id: Optional[str] = None,
        client_id: Optional[str] = None,
        api_version: Optional[str] = None,
        api_key: Optional[str] = None,
    ) -> WorkspaceConnectionParams:
        """
        Set all fields/properties with `not None` values
        passed in the (named or key-valued) arguments
        into this instance IF the instance does not have
        the corresponding parameter set yet.
        """
        self._merge(
            api_version=api_version,
            arm_endpoint=arm_endpoint,
            quantum_endpoint=quantum_endpoint,
            client_id=client_id,
            credential=credential,
            environment=environment,
            location=location,
            resource_group=resource_group,
            subscription_id=subscription_id,
            tenant_id=tenant_id,
            user_agent=user_agent,
            user_agent_app_id=user_agent_app_id,
            workspace_name=workspace_name,
            api_key=api_key,
            merge_default_mode=True,
        )
        return self

    def _merge(
        self,
        merge_default_mode: bool,
        subscription_id: Optional[str] = None,
        resource_group: Optional[str] = None,
        workspace_name: Optional[str] = None,
        location: Optional[str] = None,
        quantum_endpoint: Optional[str] = None,
        arm_endpoint: Optional[str] = None,
        environment: Union[str, EnvironmentKind, None] = None,
        credential: Optional[object] = None,
        user_agent: Optional[str] = None,
        user_agent_app_id: Optional[str] = None,
        tenant_id: Optional[str] = None,
        client_id: Optional[str] = None,
        api_version: Optional[str] = None,
        api_key: Optional[str] = None,
    ):
        """
        Set all fields/properties with `not None` values
        passed in the kwargs arguments
        into this instance.

        If merge_default_mode is True, skip setting
        the field/property if it already has a value.
        """
        def _get_value_or_default(old_value, new_value):
            if merge_default_mode and old_value:
                return old_value
            if new_value:
                return new_value
            return old_value

        self.subscription_id = _get_value_or_default(self.subscription_id, subscription_id)
        self.resource_group = _get_value_or_default(self.resource_group, resource_group)
        self.workspace_name = _get_value_or_default(self.workspace_name, workspace_name)
        self.location = _get_value_or_default(self.location, location)
        self.environment = _get_value_or_default(self.environment, environment)
        self.credential = _get_value_or_default(self.credential, credential)
        self.user_agent = _get_value_or_default(self.user_agent, user_agent)
        self.user_agent_app_id = _get_value_or_default(self.user_agent_app_id, user_agent_app_id)
        self.client_id = _get_value_or_default(self.client_id, client_id)
        self.tenant_id = _get_value_or_default(self.tenant_id, tenant_id)
        self.api_version = _get_value_or_default(self.api_version, api_version)
        self.api_key = _get_value_or_default(self.api_key, api_key)
        # for these properties that have a default value in the getter, we use
        # the private field as the old_value
        self.quantum_endpoint = _get_value_or_default(self._quantum_endpoint, quantum_endpoint)
        self.arm_endpoint = _get_value_or_default(self._arm_endpoint, arm_endpoint)
        return self

    def _merge_connection_params(
        self,
        connection_params: WorkspaceConnectionParams,
        merge_default_mode: bool = False,
    ) -> WorkspaceConnectionParams:
        """
        Set all fields/properties with `not None` values
        from the `connection_params` into this instance.
        """
        self._merge(
            api_version=connection_params.api_version,
            client_id=connection_params.client_id,
            credential=connection_params.credential,
            environment=connection_params.environment,
            location=connection_params.location,
            resource_group=connection_params.resource_group,
            subscription_id=connection_params.subscription_id,
            tenant_id=connection_params.tenant_id,
            user_agent=connection_params.user_agent,
            user_agent_app_id=connection_params.user_agent_app_id,
            workspace_name=connection_params.workspace_name,
            merge_default_mode=merge_default_mode,
            # for these properties that have a default value in the getter,
            # so we use the private field instead
            # pylint: disable=protected-access
            arm_endpoint=connection_params._arm_endpoint,
            quantum_endpoint=connection_params._quantum_endpoint,
        )
        return self

    def get_credential_or_default(self) -> Any:
        """
        Get the credential if one was set,
        or defaults to a new _DefaultAzureCredential.
        """
        return (self.credential
                or _DefaultAzureCredential(
                    subscription_id=self.subscription_id,
                    arm_endpoint=self.arm_endpoint,
                    tenant_id=self.tenant_id))

    def get_auth_policy(self) -> Any:
        """
        Returns a AzureKeyCredentialPolicy if using an AzureKeyCredential.
        Defaults to None.
        """
        if isinstance(self.credential, AzureKeyCredential):
            return AzureKeyCredentialPolicy(self.credential,
                                            ConnectionConstants.QUANTUM_API_KEY_HEADER)
        return None

    def append_user_agent(self, value: str):
        """
        Append a new value to the Workspace's UserAgent and re-initialize the
        QuantumClient. The values are appended using a dash.

        :param value: UserAgent value to add, e.g. "azure-quantum-<plugin>"
        """
        new_user_agent = None

        if (
            value
            and value not in (self.user_agent or "")
        ):
            new_user_agent = (f"{self.user_agent}-{value}"
                              if self.user_agent else value)

        if new_user_agent != self.user_agent:
            self.user_agent = new_user_agent
            if self.on_new_client_request:
                self.on_new_client_request()

    def get_full_user_agent(self):
        """
        Get the full Azure Quantum Python SDK UserAgent
        that is sent to the service via the header.
        """
        full_user_agent = self.user_agent
        app_id = self.user_agent_app_id
        if self.user_agent_app_id:
            full_user_agent = (f"{app_id} {full_user_agent}"
                               if full_user_agent else app_id)
        return full_user_agent

    def is_complete(self) -> bool:
        """
        Returns true if we have all necessary parameters
        to connect to the Azure Quantum Workspace.
        """
        return (self.location
                and self.subscription_id
                and self.resource_group
                and self.workspace_name
                and self.get_credential_or_default())

    def assert_complete(self):
        """
        Raises ValueError if we don't have all necessary parameters
        to connect to the Azure Quantum Workspace.
        """
        if not self.is_complete():
            raise ValueError(
                """
                    Azure Quantum workspace not fully specified.
                    Please specify one of the following:
                    1) A valid combination of location and resource ID.
                    2) A valid combination of location, subscription ID,
                    resource group name, and workspace name.
                    3) A valid connection string (via Workspace.from_connection_string()).
                """)

    def default_from_env_vars(self) -> WorkspaceConnectionParams:
        """
        Apply default values found in the environment variables
        if current parameters are not set.
        """
        self.subscription_id = (self.subscription_id
                                or os.environ.get(EnvironmentVariables.QUANTUM_SUBSCRIPTION_ID)
                                or os.environ.get(EnvironmentVariables.SUBSCRIPTION_ID))
        self.resource_group = (self.resource_group
                               or os.environ.get(EnvironmentVariables.QUANTUM_RESOURCE_GROUP)
                               or os.environ.get(EnvironmentVariables.RESOURCE_GROUP))
        self.workspace_name = (self.workspace_name
                               or os.environ.get(EnvironmentVariables.WORKSPACE_NAME))
        self.location = (self.location
                         or os.environ.get(EnvironmentVariables.QUANTUM_LOCATION)
                         or os.environ.get(EnvironmentVariables.LOCATION))
        self.user_agent_app_id = (self.user_agent_app_id
                                  or os.environ.get(EnvironmentVariables.USER_AGENT_APPID))
        self.tenant_id = (self.tenant_id
                          or os.environ.get(EnvironmentVariables.AZURE_TENANT_ID))
        self.client_id = (self.client_id
                          or os.environ.get(EnvironmentVariables.AZURE_CLIENT_ID))
        # for these properties we use the private field
        # because the getter return default values
        self.environment = (self._environment
                            or os.environ.get(EnvironmentVariables.QUANTUM_ENV))
        # only try to use the connection string from env var if
        # we really need it
        if (not self.location
            or not self.subscription_id
            or not self.resource_group
            or not self.workspace_name
            or not self.credential
        ):
            self._merge_connection_params(
                connection_params=WorkspaceConnectionParams(
                    connection_string=os.environ.get(EnvironmentVariables.CONNECTION_STRING)),
                merge_default_mode=True)
        return self

    @classmethod
    def from_env_vars(
        cls,
    ) -> WorkspaceConnectionParams:
        """
        Initialize the WorkspaceConnectionParams from values found
        in the environment variables.
        """
        return WorkspaceConnectionParams().default_from_env_vars()

    def _merge_re_match(self, re_match: Match[str]):
        def get_value(group_name):
            return re_match.groupdict().get(group_name)
        self.merge(
            subscription_id=get_value('subscription_id'),
            resource_group=get_value('resource_group'),
            workspace_name=get_value('workspace_name'),
            location=get_value('location'),
            quantum_endpoint=get_value('quantum_endpoint'),
            api_key=get_value('api_key'),
            arm_endpoint=get_value('arm_endpoint'),
        )



##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##


import logging
import time
import json

from typing import TYPE_CHECKING

from azure.quantum._client.models import JobDetails
from azure.quantum.job.job_failed_with_results_error import JobFailedWithResultsError
from azure.quantum.job.base_job import BaseJob, ContentType, DEFAULT_TIMEOUT
from azure.quantum.job.filtered_job import FilteredJob

__all__ = ["Job", "JobDetails"]

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from azure.quantum.workspace import Workspace


_log = logging.getLogger(__name__)


class Job(BaseJob, FilteredJob):
    """Azure Quantum Job that is submitted to a given Workspace.

    :param workspace: Workspace instance to submit job to
    :type workspace: Workspace
    :param job_details: Job details model,
            contains Job ID, name and other details
    :type job_details: JobDetails
    """

    _default_poll_wait = 0.2

    def __init__(self, workspace: "Workspace", job_details: JobDetails, **kwargs):
        self.results = None
        super().__init__(
            workspace=workspace,
            details=job_details,
            **kwargs
        )

    def submit(self):
        """Submit a job to Azure Quantum."""
        _log.debug(f"Submitting job with ID {self.id}")
        job = self.workspace.submit_job(self)
        self.details = job.details

    def refresh(self):
        """Refreshes the Job's details by querying the workspace."""
        self.details = self.workspace.get_job(self.id).details

    def has_completed(self) -> bool:
        """Check if the job has completed."""
        return (
            self.details.status == "Succeeded"
            or self.details.status == "Failed"
            or self.details.status == "Cancelled"
        )

    def wait_until_completed(
        self,
        max_poll_wait_secs=30,
        timeout_secs=None,
        print_progress=True
    ) -> None:
        """Keeps refreshing the Job's details
        until it reaches a finished status.

        :param max_poll_wait_secs: Maximum poll wait time, defaults to 30
        :type max_poll_wait_secs: int
        :param timeout_secs: Timeout in seconds, defaults to None
        :type timeout_secs: int
        :param print_progress: Print "." to stdout to display progress
        :type print_progress: bool
        :raises: :class:`TimeoutError` If the total poll time exceeds timeout, raise.
        """
        self.refresh()
        poll_wait = Job._default_poll_wait
        start_time = time.time()
        while not self.has_completed():
            if timeout_secs is not None and (time.time() - start_time) >= timeout_secs:
                raise TimeoutError(f"The wait time has exceeded {timeout_secs} seconds.")

            logger.debug(
                f"Waiting for job {self.id},"
                + f"it is in status '{self.details.status}'"
            )
            if print_progress:
                print(".", end="", flush=True)
            time.sleep(poll_wait)
            self.refresh()
            poll_wait = (
                max_poll_wait_secs
                if poll_wait >= max_poll_wait_secs
                else poll_wait * 1.5
            )

    def get_results(self, timeout_secs: float = DEFAULT_TIMEOUT):
        """Get job results by downloading the results blob from the
        storage container linked via the workspace.
        
        Raises :class:`RuntimeError` if job execution fails.
        
        Raises :class:`ValueError` if job output is malformed or output format is not compatible.

        Raises :class:`azure.quantum.job.JobFailedWithResultsError` if job execution fails, 
                but failure results could still be retrieved (e.g. for jobs submitted against "microsoft.dft" target).

        :param timeout_secs: Timeout in seconds, defaults to 300
        :type timeout_secs: float
        :return: Results dictionary with histogram shots, or raw results if not a json object.
        :rtype: typing.Any
        """
        if self.results is not None:
            return self.results

        if not self.has_completed():
            self.wait_until_completed(timeout_secs=timeout_secs)

        if not self.details.status == "Succeeded":
            if self.details.status == "Failed" and self._allow_failure_results():
                job_blob_properties = self.download_blob_properties(self.details.output_data_uri)
                if job_blob_properties.size > 0:
                    job_failure_data = self.download_data(self.details.output_data_uri)
                    raise JobFailedWithResultsError("An error occurred during job execution.", job_failure_data)

            raise RuntimeError(
                f'{"Cannot retrieve results as job execution failed"}'
                + f"(status: {self.details.status}."
                + f"error: {self.details.error_data})"
            )

        payload = self.download_data(self.details.output_data_uri)
        try:
            payload = payload.decode("utf8")
            results = json.loads(payload)

            if self.details.output_data_format == "microsoft.quantum-results.v1":
                if "Histogram" not in results:
                    raise ValueError(f"\"Histogram\" array was expected to be in the Job results for \"{self.details.output_data_format}\" output format.")
                
                histogram_values = results["Histogram"]

                if len(histogram_values) % 2 == 0:
                    # Re-mapping {'Histogram': ['[0]', 0.50, '[1]', 0.50] } to {'[0]': 0.50, '[1]': 0.50}
                    return {histogram_values[i]: histogram_values[i + 1] for i in range(0, len(histogram_values), 2)}
                else: 
                    raise ValueError(f"\"Histogram\" array has invalid format. Even number of items is expected.")
            elif self.details.output_data_format == "microsoft.quantum-results.v2":
                if "DataFormat" not in results or results["DataFormat"] != "microsoft.quantum-results.v2":
                    raise ValueError(f"\"DataFormat\" was expected to be \"microsoft.quantum-results.v2\" in the Job results for \"{self.details.output_data_format}\" output format.")

                if "Results" not in results:
                    raise ValueError(f"\"Results\" field was expected to be in the Job results for \"{self.details.output_data_format}\" output format.")
                
                if len(results["Results"]) < 1:
                    raise ValueError("\"Results\" array was expected to contain at least one item")
                
                results = results["Results"][0]

                if "Histogram" not in results:
                    raise ValueError(f"\"Histogram\" array was expected to be in the Job results for \"{self.details.output_data_format}\" output format.")

                if "Shots" not in results:
                    raise ValueError(f"\"Shots\" array was expected to be in the Job results for \"{self.details.output_data_format}\" output format.")

                histogram_values = results["Histogram"]

                total_count = len(results["Shots"])

                # Re-mapping object {'Histogram': [{"Outcome": [0], "Display": '[0]', "Count": 500}, {"Outcome": [1], "Display": '[1]', "Count": 500}]} to {'[0]': 0.50, '[1]': 0.50}
                return {outcome["Display"]: outcome["Count"] / total_count for outcome in histogram_values}
            
            return results
        except:
            # If errors decoding the data, return the raw payload:
            return payload

    def get_results_histogram(self, timeout_secs: float = DEFAULT_TIMEOUT):
        """Get job results histogram by downloading the results blob from the storage container linked via the workspace.
        
        Raises :class:`RuntimeError` if job execution fails.
        
        Raises :class:`ValueError` if job output is malformed or output format is not compatible.

        Raises :class:`azure.quantum.job.JobFailedWithResultsError` if job execution fails, 
                but failure results could still be retrieved (e.g. for jobs submitted against "microsoft.dft" target).

        :param timeout_secs: Timeout in seconds, defaults to 300
        :type timeout_secs: float
        :return: Results dictionary with histogram shots, or raw results if not a json object.
        :rtype: typing.Any
        """
        if self.results is not None:
            return self.results

        if not self.has_completed():
            self.wait_until_completed(timeout_secs=timeout_secs)

        if not self.details.status == "Succeeded":
            if self.details.status == "Failed" and self._allow_failure_results():
                job_blob_properties = self.download_blob_properties(self.details.output_data_uri)
                if job_blob_properties.size > 0:
                    job_failure_data = self.download_data(self.details.output_data_uri)
                    raise JobFailedWithResultsError("An error occurred during job execution.", job_failure_data)

            raise RuntimeError(
                f'{"Cannot retrieve results as job execution failed"}'
                + f"(status: {self.details.status}."
                + f"error: {self.details.error_data})"
            )

        payload = self.download_data(self.details.output_data_uri)
        try:
            payload = payload.decode("utf8")
            results = json.loads(payload)

            if self.details.output_data_format == "microsoft.quantum-results.v2":
                if "DataFormat" not in results or results["DataFormat"] != "microsoft.quantum-results.v2":
                    raise ValueError(f"\"DataFormat\" was expected to be \"microsoft.quantum-results.v2\" in the Job results for \"{self.details.output_data_format}\" output format.")
                if "Results" not in results:
                    raise ValueError(f"\"Results\" field was expected to be in the Job results for \"{self.details.output_data_format}\" output format.")
                
                if len(results["Results"]) < 1:
                    raise ValueError("\"Results\" array was expected to contain at least one item")
                
                results = results["Results"]

                if len(results) == 1: 
                    results = results[0]
                    if "Histogram" not in results:
                        raise ValueError(f"\"Histogram\" array was expected to be in the Job results for \"{self.details.output_data_format}\" output format.")

                    histogram_values = results["Histogram"]
                    outcome_keys = self._process_outcome(histogram_values)

                    # Re-mapping object {'Histogram': [{"Outcome": [0], "Display": '[0]', "Count": 500}, {"Outcome": [1], "Display": '[1]', "Count": 500}]} to {'[0]': {"Outcome": [0], "Count": 500}, '[1]': {"Outcome": [1], "Count": 500}}
                    return {hist_val["Display"]: {"outcome": outcome, "count": hist_val["Count"]} for outcome, hist_val in zip(outcome_keys, histogram_values)}

                else:
                    # This is handling the BatchResults edge case
                    resultsArray = []
                    for i, result in enumerate(results):
                        if "Histogram" not in result:
                            raise ValueError(f"\"Histogram\" array was expected to be in the Job results for result {i} for \"{self.details.output_data_format}\" output format.")

                        histogram_values = result["Histogram"]
                        outcome_keys = self._process_outcome(histogram_values)

                        # Re-mapping object {'Histogram': [{"Outcome": [0], "Display": '[0]', "Count": 500}, {"Outcome": [1], "Display": '[1]', "Count": 500}]} to {'[0]': {"Outcome": [0], "Count": 500}, '[1]': {"Outcome": [1], "Count": 500}}
                        resultsArray.append({hist_val["Display"]: {"outcome": outcome, "count": hist_val["Count"]} for outcome, hist_val in zip(outcome_keys, histogram_values)})

                    return resultsArray

            else:
                raise ValueError(f"Getting a results histogram with counts instead of probabilities is not a supported feature for jobs using the \"{self.details.output_data_format}\" output format.")

        except Exception as e:
            raise e

    def get_results_shots(self, timeout_secs: float = DEFAULT_TIMEOUT):
        """Get job results per shot data by downloading the results blob from the
        storage container linked via the workspace.
        
        Raises :class:`RuntimeError` if job execution fails.
        
        Raises :class:`ValueError` if job output is malformed or output format is not compatible.

        Raises :class:`azure.quantum.job.JobFailedWithResultsError` if job execution fails, 
                but failure results could still be retrieved (e.g. for jobs submitted against "microsoft.dft" target).

        :param timeout_secs: Timeout in seconds, defaults to 300
        :type timeout_secs: float
        :return: Results dictionary with histogram shots, or raw results if not a json object.
        :rtype: typing.Any
        """
        if self.results is not None:
            return self.results

        if not self.has_completed():
            self.wait_until_completed(timeout_secs=timeout_secs)

        if not self.details.status == "Succeeded":
            if self.details.status == "Failed" and self._allow_failure_results():
                job_blob_properties = self.download_blob_properties(self.details.output_data_uri)
                if job_blob_properties.size > 0:
                    job_failure_data = self.download_data(self.details.output_data_uri)
                    raise JobFailedWithResultsError("An error occurred during job execution.", job_failure_data)

            raise RuntimeError(
                f'{"Cannot retrieve results as job execution failed"}'
                + f"(status: {self.details.status}."
                + f"error: {self.details.error_data})"
            )

        payload = self.download_data(self.details.output_data_uri)
        try:
            payload = payload.decode("utf8")
            results = json.loads(payload)

            if self.details.output_data_format == "microsoft.quantum-results.v2":
                if "DataFormat" not in results or results["DataFormat"] != "microsoft.quantum-results.v2":
                    raise ValueError(f"\"DataFormat\" was expected to be \"microsoft.quantum-results.v2\" in the Job results for \"{self.details.output_data_format}\" output format.")
                if "Results" not in results:
                    raise ValueError(f"\"Results\" field was expected to be in the Job results for \"{self.details.output_data_format}\" output format.")

                results = results["Results"]

                if len(results) < 1:
                    raise ValueError("\"Results\" array was expected to contain at least one item")

                if len(results) == 1: 
                    result = results[0]
                    if "Shots" not in result:
                        raise ValueError(f"\"Shots\" array was expected to be in the Job results for \"{self.details.output_data_format}\" output format.")
                    
                    return [self._convert_tuples(shot) for shot in result["Shots"]]
                else:
                    # This is handling the BatchResults edge case
                    shotsArray = []
                    for i, result in enumerate(results):
                        if "Shots" not in result:
                            raise ValueError(f"\"Shots\" array was expected to be in the Job results for result {i} of \"{self.details.output_data_format}\" output format.")
                        shotsArray.append([self._convert_tuples(shot) for shot in result["Shots"]])
                    
                    return shotsArray
            else:   
                raise ValueError(f"Individual shot results are not supported for jobs using the \"{self.details.output_data_format}\" output format.")
        except Exception as e:
            raise e

    def _process_outcome(self, histogram_results):
        return [self._convert_tuples(v['Outcome']) for v in histogram_results]

    def _convert_tuples(self, data):
        if isinstance(data, dict):
            # Check if the dictionary represents a tuple
            if all(isinstance(k, str) and k.startswith("Item") for k in data.keys()):
                # Convert the dictionary to a tuple
                return tuple(self._convert_tuples(data[f"Item{i+1}"]) for i in range(len(data)))
            else:
                raise "Malformed tuple output"
        elif isinstance(data, list):
            # Recursively process list elements
            return [self._convert_tuples(item) for item in data]
        else:
            # Return the data as is (int, string, etc.)
            return data

    @classmethod
    def _allow_failure_results(cls) -> bool: 
        """
        Allow to download job results even if the Job status is "Failed".

        This method can be overridden in derived classes to alter the default
        behaviour.

        The default is False.
        """
        return False


##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##

from enum import Enum
from azure.identity._constants import EnvironmentVariables as SdkEnvironmentVariables
from azure.identity import _internal as AzureIdentityInternals


class EnvironmentVariables:
    USER_AGENT_APPID = "AZURE_QUANTUM_PYTHON_APPID"
    QUANTUM_LOCATION = "AZURE_QUANTUM_WORKSPACE_LOCATION"
    LOCATION = "LOCATION"
    QUANTUM_RESOURCE_GROUP = "AZURE_QUANTUM_WORKSPACE_RG"
    RESOURCE_GROUP = "RESOURCE_GROUP"
    QUANTUM_SUBSCRIPTION_ID = "AZURE_QUANTUM_SUBSCRIPTION_ID"
    SUBSCRIPTION_ID = "SUBSCRIPTION_ID"
    WORKSPACE_NAME = "AZURE_QUANTUM_WORKSPACE_NAME"
    QUANTUM_ENV = "AZURE_QUANTUM_ENV"
    AZURE_CLIENT_ID = SdkEnvironmentVariables.AZURE_CLIENT_ID
    AZURE_CLIENT_SECRET = SdkEnvironmentVariables.AZURE_CLIENT_SECRET
    AZURE_CLIENT_CERTIFICATE_PATH = SdkEnvironmentVariables.AZURE_CLIENT_CERTIFICATE_PATH
    AZURE_CLIENT_SEND_CERTIFICATE_CHAIN = SdkEnvironmentVariables.AZURE_CLIENT_SEND_CERTIFICATE_CHAIN
    AZURE_TENANT_ID = SdkEnvironmentVariables.AZURE_TENANT_ID
    QUANTUM_TOKEN_FILE = "AZURE_QUANTUM_TOKEN_FILE"
    CONNECTION_STRING = "AZURE_QUANTUM_CONNECTION_STRING"
    ALL = [
        USER_AGENT_APPID,
        QUANTUM_LOCATION,
        LOCATION,
        QUANTUM_RESOURCE_GROUP,
        RESOURCE_GROUP,
        QUANTUM_SUBSCRIPTION_ID,
        SUBSCRIPTION_ID,
        WORKSPACE_NAME,
        QUANTUM_ENV,
        AZURE_CLIENT_ID,
        AZURE_CLIENT_SECRET,
        AZURE_CLIENT_CERTIFICATE_PATH,
        AZURE_CLIENT_SEND_CERTIFICATE_CHAIN,
        AZURE_TENANT_ID,
        QUANTUM_TOKEN_FILE,
        CONNECTION_STRING,
    ]


class EnvironmentKind(Enum):
    PRODUCTION = 1,
    CANARY = 2,
    DOGFOOD = 3


class ConnectionConstants:
    DATA_PLANE_CREDENTIAL_SCOPE = "https://quantum.microsoft.com/.default"
    ARM_CREDENTIAL_SCOPE = "https://management.azure.com/.default"

    MSA_TENANT_ID = "9188040d-6c67-4c5b-b112-36a304b66dad"

    AUTHORITY = AzureIdentityInternals.get_default_authority()
    DOGFOOD_AUTHORITY = "login.windows-ppe.net"

    # pylint: disable=unnecessary-lambda-assignment
    GET_QUANTUM_PRODUCTION_ENDPOINT = \
        lambda location: f"https://{location}.quantum.azure.com/"
    GET_QUANTUM_CANARY_ENDPOINT = \
        lambda location: f"https://{location or 'eastus2euap'}.quantum.azure.com/"
    GET_QUANTUM_DOGFOOD_ENDPOINT = \
        lambda location: f"https://{location}.quantum-test.azure.com/"

    ARM_PRODUCTION_ENDPOINT = "https://management.azure.com/"
    ARM_DOGFOOD_ENDPOINT = "https://api-dogfood.resources.windows-int.net/"

    VALID_RESOURCE_ID = (
        lambda subscription_id, resource_group, workspace_name:
        f"/subscriptions/{subscription_id}" +
        f"/resourceGroups/{resource_group}" +
        "/providers/Microsoft.Quantum/" +
        f"Workspaces/{workspace_name}"
    )

    VALID_CONNECTION_STRING = (
        lambda subscription_id, resource_group, workspace_name, api_key, quantum_endpoint:
        f"SubscriptionId={subscription_id};" +
        f"ResourceGroupName={resource_group};" +
        f"WorkspaceName={workspace_name};" +
        f"ApiKey={api_key};" +
        f"QuantumEndpoint={quantum_endpoint};"
    )

    QUANTUM_API_KEY_HEADER = "x-ms-quantum-api-key"

GUID_REGEX_PATTERN = (
    r"[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}"
)
